@@ -1,10 +1,9 @@
 from typing import Union
 
-import casadi
 from casadi import Function, vertcat, horzcat, norm_fro, collocation_points, tangent, rootfinder, MX, SX
 import numpy as np
 
-from ..misc.enums import ControlType, OdeSolver
+from ..misc.enums import ControlType
 
 
 class Integrator:
@@ -19,7 +18,7 @@ class Integrator:
         The initial and final time
     idx: int
         The index of the degrees of freedom to integrate
-    CX: Union[MX, SX]
+    cx: Union[MX, SX]
         The CasADi type the integration should be built from
     x_sym: Union[MX, SX]
         The state variables
@@ -27,7 +26,9 @@ class Integrator:
         The control variables
     param_sym: Union[MX, SX]
         The parameters variables
-    fun: function
+    param_scaling: Union[MX, SX]
+        The parameters variables scaling factor
+    fun: Callable
         The dynamic function which provides the derivative of the states
     control_type: ControlType
         The type of the controls
@@ -40,6 +41,8 @@ class Integrator:
 
     Methods
     -------
+    __call__(self, *args, **kwargs)
+        Interface to self.function
     map(self, *args, **kwargs) -> Function
         Get the multithreaded CasADi graph of the integration
     get_u(self, u: np.ndarray, dt_norm: float) -> np.ndarray
@@ -48,8 +51,6 @@ class Integrator:
         The dynamics of the system
     _finish_init(self)
         Prepare the CasADi function from dxdt
-    prepare_dynamic_integrator(ocp: OptimalControlProgram, nlp: NonLinearProgram)
-        Properly set the integration in an nlp
     """
 
     # Todo change ode and ode_opt into class
@@ -62,13 +63,15 @@ class Integrator:
         ode_opt: dict
             The ode options
         """
+
         self.model = ode_opt["model"]
         self.t_span = ode_opt["t0"], ode_opt["tf"]
         self.idx = ode_opt["idx"]
-        self.CX = ode_opt["CX"]
+        self.cx = ode_opt["cx"]
         self.x_sym = ode["x"]
         self.u_sym = ode["p"]
-        self.param_sym = ode_opt["param"]
+        self.param_sym = ode_opt["param"].cx
+        self.param_scaling = ode_opt["param"].scaling
         self.fun = ode["ode"]
         self.control_type = ode_opt["control_type"]
         self.step_time = self.t_span[1] - self.t_span[0]
@@ -76,6 +79,10 @@ class Integrator:
         self.function = None
 
     def __call__(self, *args, **kwargs):
+        """
+        Interface to self.function
+        """
+
         return self.function(*args, **kwargs)
 
     def map(self, *args, **kwargs) -> Function:
@@ -141,87 +148,11 @@ class Integrator:
         self.function = Function(
             "integrator",
             [self.x_sym, self.u_sym, self.param_sym],
-            self.dxdt(self.h, self.x_sym, self.u_sym, self.param_sym),
+            self.dxdt(self.h, self.x_sym, self.u_sym, self.param_sym * self.param_scaling),
             ["x0", "p", "params"],
             ["xf", "xall"],
         )
 
-    @staticmethod
-    def prepare_dynamic_integrator(ocp, nlp):
-        """
-        Properly set the integration in an nlp
-
-        Parameters
-        ----------
-        ocp: OptimalControlProgram
-            A reference to the main program
-        nlp: NonLinearProgram
-            A reference to the current phase of the ocp
-        """
-
-        ode_opt = {"t0": 0, "tf": nlp.dt}
-        if nlp.ode_solver == OdeSolver.RK4 or nlp.ode_solver == OdeSolver.RK8:
-            ode_opt["number_of_finite_elements"] = nlp.n_integration_steps
-        elif nlp.ode_solver == OdeSolver.IRK:
-            nlp.n_integration_steps = 1
-
-        dynamics = nlp.dynamics_func
-        ode = {"x": nlp.x, "p": nlp.u, "ode": dynamics(nlp.x, nlp.u, nlp.p)}
-        nlp.dynamics = []
-        nlp.par_dynamics = {}
-        if nlp.ode_solver == OdeSolver.RK4 or nlp.ode_solver == OdeSolver.RK8 or nlp.ode_solver == OdeSolver.IRK:
-            if nlp.ode_solver == OdeSolver.IRK:
-                if ocp.CX is SX:
-                    raise NotImplementedError("use_sx and OdeSolver.IRK are not yet compatible")
-
-                if nlp.model.nbQuat() > 0:
-                    raise NotImplementedError(
-                        "Quaternions can't be used with IRK yet. If you get this error, please notify the "
-                        "developers and ping EveCharbie"
-                    )
-
-            ode_opt["model"] = nlp.model
-            ode_opt["param"] = nlp.p
-            ode_opt["CX"] = nlp.CX
-            ode_opt["idx"] = 0
-            ode["ode"] = dynamics
-            ode_opt["control_type"] = nlp.control_type
-            if nlp.external_forces:
-                for idx in range(len(nlp.external_forces)):
-                    ode_opt["idx"] = idx
-                    if nlp.ode_solver == OdeSolver.RK4:
-                        nlp.dynamics.append(RK4(ode, ode_opt))
-                    elif nlp.ode_solver == OdeSolver.RK8:
-                        nlp.dynamics.append(RK8(ode, ode_opt))
-                    elif nlp.ode_solver == OdeSolver.IRK:
-                        ode_opt["irk_polynomial_interpolation_degree"] = nlp.irk_polynomial_interpolation_degree
-                        nlp.dynamics.append(IRK(ode, ode_opt))
-            else:
-                if ocp.n_threads > 1 and nlp.control_type == ControlType.LINEAR_CONTINUOUS:
-                    raise RuntimeError("Piece-wise linear continuous controls cannot be used with multiple threads")
-                if nlp.ode_solver == OdeSolver.RK4:
-                    nlp.dynamics.append(RK4(ode, ode_opt))
-                if nlp.ode_solver == OdeSolver.RK8:
-                    nlp.dynamics.append(RK8(ode, ode_opt))
-                elif nlp.ode_solver == OdeSolver.IRK:
-                    ode_opt["irk_polynomial_interpolation_degree"] = nlp.irk_polynomial_interpolation_degree
-                    nlp.dynamics.append(IRK(ode, ode_opt))
-        elif nlp.ode_solver == OdeSolver.CVODES:
-            if not isinstance(ocp.CX(), MX):
-                raise RuntimeError("CVODES integrator can only be used with MX graphs")
-            if len(ocp.param_to_optimize) != 0:
-                raise RuntimeError("CVODES cannot be used while optimizing parameters")
-            if nlp.external_forces:
-                raise RuntimeError("CVODES cannot be used with external_forces")
-            if nlp.control_type == ControlType.LINEAR_CONTINUOUS:
-                raise RuntimeError("CVODES cannot be used with piece-wise linear controls (only RK4)")
-            nlp.dynamics.append(casadi.integrator("integrator", "cvodes", ode, ode_opt))
-
-        if len(nlp.dynamics) == 1:
-            if ocp.n_threads > 1:
-                nlp.par_dynamics = nlp.dynamics[0].map(nlp.ns, "thread", ocp.n_threads)
-            nlp.dynamics = nlp.dynamics * nlp.ns
-
 
 class RK(Integrator):
     """
@@ -233,6 +164,8 @@ class RK(Integrator):
         Number of finite element during the integration
     h_norm: float
         Normalized time step
+    h: float
+        Length of steps
 
     Methods
     -------
@@ -255,7 +188,7 @@ class RK(Integrator):
         super(RK, self).__init__(ode, ode_opt)
         self.n_step = ode_opt["number_of_finite_elements"]
         self.h_norm = 1 / self.n_step
-        self.h = self.step_time * self.h_norm  # Length of steps
+        self.h = self.step_time * self.h_norm
 
     def next_x(self, h: float, t: float, x_prev: Union[MX, SX], u: Union[MX, SX], p: Union[MX, SX]):
         """
@@ -302,7 +235,7 @@ class RK(Integrator):
         """
 
         u = controls
-        x = self.CX(states.shape[0], self.n_step + 1)
+        x = self.cx(states.shape[0], self.n_step + 1)
         p = params
         x[:, 0] = states
 
@@ -460,7 +393,7 @@ class RK8(RK4):
         return x_prev + h / 840 * (41 * k1 + 27 * k4 + 272 * k5 + 27 * k6 + 216 * k7 + 216 * k9 + 41 * k10)
 
 
-class IRK(Integrator):
+class COLLOCATION(Integrator):
     """
     Numerical integration using implicit Runge-Kutta method.
 
@@ -487,8 +420,47 @@ class IRK(Integrator):
             The ode options
         """
 
-        super(IRK, self).__init__(ode, ode_opt)
+        super(COLLOCATION, self).__init__(ode, ode_opt)
+
+        self.method = ode_opt["method"]
         self.degree = ode_opt["irk_polynomial_interpolation_degree"]
+
+        # Coefficients of the collocation equation
+        self._c = self.cx.zeros((self.degree + 1, self.degree + 1))
+
+        # Coefficients of the continuity equation
+        self._d = self.cx.zeros(self.degree + 1)
+
+        # Choose collocation points
+        time_points = [0] + collocation_points(self.degree, self.method)
+
+        # Dimensionless time inside one control interval
+        time_control_interval = self.cx.sym("time_control_interval")
+
+        # For all collocation points
+        for j in range(self.degree + 1):
+            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
+            _l = 1
+            for r in range(self.degree + 1):
+                if r != j:
+                    _l *= (time_control_interval - time_points[r]) / (time_points[j] - time_points[r])
+
+            # Evaluate the polynomial at the final time to get the coefficients of the continuity equation
+            lfcn = Function("lfcn", [time_control_interval], [_l])
+            self._d[j] = lfcn(1.0)
+
+            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
+            _l = 1
+            for r in range(self.degree + 1):
+                if r != j:
+                    _l *= (time_control_interval - time_points[r]) / (time_points[j] - time_points[r])
+
+            # Evaluate the time derivative of the polynomial at all collocation points to get
+            # the coefficients of the continuity equation
+            tfcn = Function("tfcn", [time_control_interval], [tangent(_l, time_control_interval)])
+            for r in range(self.degree + 1):
+                self._c[j, r] = tfcn(time_points[r])
+
         self._finish_init()
 
     def get_u(self, u: np.ndarray, dt_norm: float) -> np.ndarray:
@@ -508,7 +480,7 @@ class IRK(Integrator):
         """
 
         if self.control_type == ControlType.CONSTANT:
-            return super(IRK, self).get_u(u, dt_norm)
+            return super(COLLOCATION, self).get_u(u, dt_norm)
         else:
             raise NotImplementedError(f"{self.control_type} ControlType not implemented yet with IRK")
 
@@ -532,74 +504,113 @@ class IRK(Integrator):
         The derivative of the states
         """
 
-        nu = controls.shape[0]
-        nx = states.shape[0]
-
-        # Choose collocation points
-        time_points = [0] + collocation_points(self.degree, "legendre")
-
-        # Coefficients of the collocation equation
-        C = self.CX.zeros((self.degree + 1, self.degree + 1))
-
-        # Coefficients of the continuity equation
-        D = self.CX.zeros(self.degree + 1)
-
-        # Dimensionless time inside one control interval
-        time_control_interval = self.CX.sym("time_control_interval")
-
-        # For all collocation points
-        for j in range(self.degree + 1):
-            # Construct Lagrange polynomials to get the polynomial basis at the collocation point
-            L = 1
-            for r in range(self.degree + 1):
-                if r != j:
-                    L *= (time_control_interval - time_points[r]) / (time_points[j] - time_points[r])
-
-            # Evaluate the polynomial at the final time to get the coefficients of the continuity equation
-            lfcn = Function("lfcn", [time_control_interval], [L])
-            D[j] = lfcn(1.0)
-
-            # Evaluate the time derivative of the polynomial at all collocation points to get
-            # the coefficients of the continuity equation
-            tfcn = Function("tfcn", [time_control_interval], [tangent(L, time_control_interval)])
-            for r in range(self.degree + 1):
-                C[j, r] = tfcn(time_points[r])
-
         # Total number of variables for one finite element
-        x0 = states
-        u = controls
-
-        x_irk_points = [self.CX.sym(f"X_irk_{j}", nx, 1) for j in range(1, self.degree + 1)]
-        x = [x0] + x_irk_points
-
-        x_irk_points_eq = []
+        states_end = self._d[0] * states[0]
+        defects = []
         for j in range(1, self.degree + 1):
 
             t_norm_init = (j - 1) / self.degree  # normalized time
             # Expression for the state derivative at the collocation point
             xp_j = 0
             for r in range(self.degree + 1):
-                xp_j += C[r, j] * x[r]
+                xp_j += self._c[r, j] * states[r]
 
             # Append collocation equations
-            f_j = self.fun(x[j], self.get_u(u, t_norm_init), params)[:, self.idx]
-            x_irk_points_eq.append(h * f_j - xp_j)
+            f_j = self.fun(states[j], self.get_u(controls, t_norm_init), params)[:, self.idx]
+            defects.append(h * f_j - xp_j)
+
+            # Add contribution to the end state
+            states_end = states_end + self._d[j] * states[j - 1]
 
         # Concatenate constraints
-        x_irk_points = vertcat(*x_irk_points)
-        x_irk_points_eq = vertcat(*x_irk_points_eq)
+        defects = vertcat(*defects)
+        return states_end, horzcat(states[0], states_end), defects
 
-        # Root-finding function, implicitly defines x_irk_points as a function of x0 and p
-        vfcn = Function("vfcn", [x_irk_points, x0, u, params], [x_irk_points_eq]).expand()
+    def _finish_init(self):
+        """
+        Prepare the CasADi function from dxdt
+        """
+
+        self.function = Function(
+            "integrator",
+            [horzcat(*self.x_sym), self.u_sym, self.param_sym],
+            self.dxdt(self.h, self.x_sym, self.u_sym, self.param_sym * self.param_scaling),
+            ["x0", "p", "params"],
+            ["xf", "xall", "defects"],
+        )
+
+
+class IRK(COLLOCATION):
+    """
+    Numerical integration using implicit Runge-Kutta method.
+
+    Methods
+    -------
+    get_u(self, u: np.ndarray, dt_norm: float) -> np.ndarray
+        Get the control at a given time
+    dxdt(self, h: float, states: Union[MX, SX], controls: Union[MX, SX], params: Union[MX, SX]) -> tuple[SX, list[SX]]
+        The dynamics of the system
+    """
+
+    def __init__(self, ode: dict, ode_opt: dict):
+        """
+        Parameters
+        ----------
+        ode: dict
+            The ode description
+        ode_opt: dict
+            The ode options
+        """
+
+        super(IRK, self).__init__(ode, ode_opt)
+
+    def dxdt(self, h: float, states: Union[MX, SX], controls: Union[MX, SX], params: Union[MX, SX]) -> tuple:
+        """
+        The dynamics of the system
+
+        Parameters
+        ----------
+        h: float
+            The time step
+        states: Union[MX, SX]
+            The states of the system
+        controls: Union[MX, SX]
+            The controls of the system
+        params: Union[MX, SX]
+            The parameters of the system
+
+        Returns
+        -------
+        The derivative of the states
+        """
+
+        nx = states[0].shape[0]
+        _, _, defect = super(IRK, self).dxdt(h, states, controls, params)
+
+        # Root-finding function, implicitly defines x_collocation_points as a function of x0 and p
+        vfcn = Function("vfcn", [vertcat(*states[1:]), states[0], controls, params], [defect]).expand()
 
         # Create a implicit function instance to solve the system of equations
         ifcn = rootfinder("ifcn", "newton", vfcn)
-        x_irk_points = ifcn(self.CX(), x0, u, params)
-        x = [x0 if r == 0 else x_irk_points[(r - 1) * nx : r * nx] for r in range(self.degree + 1)]
+        x_irk_points = ifcn(self.cx(), states[0], controls, params)
+        x = [states[0] if r == 0 else x_irk_points[(r - 1) * nx : r * nx] for r in range(self.degree + 1)]
 
         # Get an expression for the state at the end of the finite element
-        xf = self.CX.zeros(nx, self.degree + 1)  # 0 #
+        xf = self.cx.zeros(nx, self.degree + 1)  # 0 #
         for r in range(self.degree + 1):
-            xf[:, r] = xf[:, r - 1] + D[r] * x[r]
+            xf[:, r] = xf[:, r - 1] + self._d[r] * x[r]
 
-        return xf[:, -1], horzcat(x0, xf[:, -1])
+        return xf[:, -1], horzcat(states[0], xf[:, -1])
+
+    def _finish_init(self):
+        """
+        Prepare the CasADi function from dxdt
+        """
+
+        self.function = Function(
+            "integrator",
+            [self.x_sym[0], self.u_sym, self.param_sym],
+            self.dxdt(self.h, self.x_sym, self.u_sym, self.param_sym * self.param_scaling),
+            ["x0", "p", "params"],
+            ["xf", "xall"],
+        )
